# -*- coding: utf-8 -*-
#! python3

# Structure Arpèges:
# A partir de l'entrée du catalogue (offset absolu 0x40) trouver l'entée catalogue DARP (4 octets), les 4 octets suivants sont l'offset absolu de l'entrée datas DARP
# A cet offset, noté OD pour OffsetDARP on trouve la chaine 'DARP' suivi de la longueur totale des données d'arpèges, noté LA pour LenArpéges.
#
# A la fin des données d'arpèges se trouve un tableau de 256 éléments de 8 octets. Structure :
# 4 octets : offset (à patir l'offset  OD + 0x61C) des données de l'arpège dans la liste chainée (description ci-après).
# 4 octets : taille des données (big endian) de l'arpège dans la liste chainée. Si taille = 0 pas de données d'arpèges (*)
# On calcule l'offset du début du tableau : OD + LA + 8 - 256*8
#
# (*) Il peut y avoir des trous dans la liste d'arpèges. (Ex : on peut créer sa première arpège avec le numéro 256)
#
# Ce tableau est le point d'entrée de la liste d'arpèges :
# Pour chaque élément du tableau (256 arpèges) on regarde si la taille est différente de 0
# Si oui : on parcourt la liste chainée à partir de l'offset du début de l'arpège pour la taille donnée et on récupère les 6 octets d'un tableau (voir après) à l' offset OD + 0x1C + (num_arp * 6) . Avec num_arp : 0..255
#
# On trouve à l'offset OD + 0x1C un tableau de 256 éléments de 6 octets de structure suivante :
# 1 octet = 0. 1 octet = num d'arpège. 1 octet = 0. 1 octet = num d'arpège. 1 octet = catégorie d'arpège (FF = pas de type). 1 octet = sous catégorie d'arpège (FF = pas de sous type)
#
# On trouve ensuite, à l'offset OD + 0x61C, une liste chainée de chunks de structure suivante :
# 4 octets : Nom du chunk. 4 octets: taille des datas du chunk sur 4 octets (big endian) = nd. nd octets : datas du chunk.
#
# Noms des chunks:
# FFFF : Fin de la liste chainée des arpèges
# PHhd : début de l'arpège.
# Structure des datas du chunk PHhd : 6 octets puis nom de l'arpège  (ex :00.01.00.nb_tracks.01.e0.nom_de_l_arpege)
# (nb) La taille de ce chunk est variable et la longueur du nom peut-être inférieure à 20, arpèges provenant de librairie tierce.
# ATrk : chunk de piste ( 1 à 4 chunk de piste pour 1 arpège)
# ARPa : chunk global de l'arpège.
# Structure des datas du chunk ARPa : 4 octets : len du nom (big endian). nom (offset 4) + reste des datas (27 octets ?)
# TRKa : chunk provenant d'arpèges de librairie tierce. Signification inconnue
# PHRa : chunk provenant d'arpèges de librairie tierce. Signification inconnue

# Offsets directs pour un fichier .x6g (version 1.0.4):
# tableau1 (header): 0xA1
# liste chainée (datas) : 0x6A1
# tableau2 (footer) : 0x1006A1

import os.path
import sys

def bg(bigendian):
    bg = int(bigendian[3]) + int(bigendian[2]) * 0x100 + int(bigendian[1]) * 0x10000 + int(bigendian[0]) * 0x1000000
    return bg

def findOffsetDARP(datas):
    version = datas[0x10:0x10+5]
    if version != b'1.0.3' and version != b'1.0.4':
        input('Version du fichier non supportée (%s). Presser une touche pour quitter.'% version.decode())
        exit()
    lenCatalogue = bg(datas[0x20:0x20+4])
    catalogue = datas[0x40:0x40+lenCatalogue]
    cursor = 0
    trouve = False
    od = 0
    while cursor < lenCatalogue:
        if catalogue[cursor: cursor+4] == b'DARP':
            trouve = True
            od = bg(catalogue[cursor+4:cursor+8])
        cursor += 8
    if trouve == False:
        input ('Entree Catalogue DARP non trouvée. Presser une touche pour quitter.')
        exit()
    return od

def litDirArp(datas, od):
    offsetfin = bg(datas[od+4:od+8])
    delta = 256*8
    offsetdebut = offsetfin - delta
    cursor = od + 8 + offsetdebut
    arpdir = []
    for i in range(256):
        offsetArp = bg(datas[cursor: cursor+4])
        lenArp = bg(datas[cursor+4: cursor+8])
        cursor += 8
        if lenArp > 0 :
            arpdir.append([i, offsetArp, lenArp])
    return arpdir

def getDatasArp(datas, od, arpdir):
    datasarp = []
    for e in arpdir:
        offset = e[1]
        lenarp = e[2]
        datasarp.append(datas[od + 0x61C + offset : od + 0x61C + offset + lenarp])
    return datasarp

def getNom(arpdata):
    lenChunk1 = bg(arpdata[4:8])
    lenNom = lenChunk1 - 6
    nom = arpdata[4+4+6: 4+4+6+lenNom]
    return nom.decode()

def getCategorie(n, datas, od):
    offset = od + 0x1C + (n * 6)
    categorie=[]
    for i in range(6):
        categorie.append(datas[offset+i])
    if categorie[0] != 0 or categorie[2] != 0:
        input('Format Categorie inconnu. Presser une touche pour quitter.')
        exit()
    return categorie


def deChunkArp(arpdata, lenarp):
    chunks = []
    cursor = 0
    while cursor < lenarp:
        nomchunk = arpdata[cursor:cursor+4]
        lenchunk = bg(arpdata[cursor+4: cursor+8])
        datachunk = arpdata[cursor+8: cursor+8+lenchunk]
        chunks.append([nomchunk, lenchunk, datachunk])
        cursor += 8 + lenchunk
    return chunks

def printArpInfo(nomFic, datas, od, arpdir, datasarp):
    print('Fichier:%s' % nomFic)
    for x, e in enumerate(arpdir):
        arpdata = datasarp[x]
        nom = getNom(arpdata)
        categorie = getCategorie(e[0], datas, od)
        print('%03d num:%03d Offset:0x%08X Len:0x%08X Type:0x%02X SousType:0x%02X Nom:%s.' % (x, e[0], e[1], e[2], categorie[4], categorie[5], nom))
    nbarp = len(arpdir)
    print('%d Arpèges' % nbarp)
    print('OD=0x%02X' % od)

def pg1():  # Fonction interactive
    global n1, datas1, od1, arpdir1, datasarp1
    printArpInfo(n1, datas1, od1, arpdir1, datasarp1)

def pg2():  # Fonction interactive
    global n2, datas2, od2, arpdir2, datasarp2
    printArpInfo(n2, datas2, od2, arpdir2, datasarp2)

def pg3():  # Fonction interactive
    global n3, datas3, od3, arpdir3, datasarp3
    printArpInfo(n3, datas3, od3, arpdir3, datasarp3)

def printArp(n, datasarp, arpdir, datas, od):
    dirarp = arpdir[n]
    arpdata = datasarp[n]
    nom = getNom(arpdata)
    categorie = getCategorie(dirarp[0], datas, od)
    print('Num Arpège:%03d Type:0x%02X SousType:0x%02X Offset:0x%08X Len:0x%08X Nom:%s.' % (dirarp[0], categorie[4], categorie[5], dirarp[1],dirarp[2], nom))
    arpchunks = deChunkArp(arpdata, dirarp[2])
    for chunk in arpchunks:
        print('Chunk:%s Len:0x%04X  Datas:' % (chunk[0].decode(), chunk[1]))
        s = ''
        for e in chunk[2]:
            s += '%02X:' % int(e)
        print(s)

def p1(n):  # Fonction interactive
    global datasarp1, arpdir1, datas1, od1
    printArp(n, datasarp1, arpdir1, datas1, od1)

def p2(n): # Fonction interactive
    global datasarp2, arpdir2, datas2, od2
    printArp(n, datasarp2, arpdir2, datas2, od2)

def writeinfo(nomfic, arpdir, datasarp, datas, od):
    (filepath, filename) = os.path.split(nomfic)
    (shortname, extension) = os.path.splitext(filename)
    nomOut = filepath + shortname +'.txt'
    ficOut = open(nomOut, 'w')
    print('Fichier:%s' % nomfic, file = ficOut)
    for x, e in enumerate(arpdir):
        arpdata = datasarp[x]
        nom = getNom(arpdata)
        categorie = getCategorie(e[0], datas, od)
        print('%03d num:%03d Type:0x%02X SousType:0x%02X Nom:%s.' % (x, e[0], categorie[4], categorie[5], nom), file = ficOut)
    nbarp = len(arpdir)
    print('%d Arpèges' % nbarp, file = ficOut)
    ficOut.close()

def writeArp(nomOut, datas):
    ficOut = open(nomOut, 'wb')
    ficOut.write(datas)
    ficOut.close()

def expandStringList(s, maxLeft, maxRight):
    # Format liste 'a2,b4,a3..12,b8,b15..20' lettre_liste: a..b , num_arpeges: 1..256
    listeMerge = []
    t = s.split(',')
    for e in t:
        x = e.find('..')
        if x >= 0:
            iliste = ord(e[0]) - ord('a')
            if not(-1 < iliste < 2):
                input('Lettre de liste invalide. Presser une touche pour quitter.')
                exit()
            imin = int(e[1:x])
            imax = int(e[x+2:])
            lmax = maxLeft if iliste == 0 else maxRight
            if imin > lmax or imin < 1 or imax > lmax or imax < 1 or imax < imin:
                input("Numéro d'arpège invalide.  Presser une touche pour quitter.")
                exit()
            for i in range(imin, imax+1):
                listeMerge.append([iliste, i])
        else:
            iliste = ord(e[0]) - ord('a')
            if not(-1 < iliste < 2):
                input('Lettre de liste invalide. Presser une touche pour quitter.')
                exit()
            num = int(e[1:])
            lmax = maxLeft if iliste == 0 else maxRight
            if num > lmax or num < 1:
                input("Numéro d'arpège invalide.  Presser une touche pour quitter.")
                exit()
            listeMerge.append([iliste, num])
    return listeMerge

def makeEmptyArps():
    s = b'YAMAHA-YSFC'
    s += b'\x00' * 5
    s += b'1.0.4'
    s += b'\x00' * 14
    s += b'\x10'
    s += b'\xff' * 28
    s += b'EARP\x00\x00\x00PDARP\x00\x00\x00\x85'
    s += b'\x45\x41\x52\x50\x00\x00\x00\x2d\x00\x00\x00\x01\x45\x6e\x74\x72'
    s += b'\x00\x00\x00\x21\x00\x10\x0e\x08\x00\x00\x00\x0c\x00\x00\x00\x00'
    s += b'\x55\x73\x65\x72\x20\x41\x72\x70\x65\x67\x67\x69\x6f\x20\x20\x20'
    s += b'\x20\x20\x20\x20\x00\x44\x41\x52\x50\x00\x10\x0e\x14\x00\x00\x00'
    s += b'\x01\x44\x61\x74\x61\x00\x10\x0e\x08\x0d\xa9\x76\xe0\x00\x00\x01\x00'
    for i in range(256):
        s += i.to_bytes(2, 'big')
        s += i.to_bytes(2, 'big')
        s += b'\xff\xff'
    s += b'\xff' * 1048576
    s += b'\x00' * 256 * 8
    return s

def findOffsetListChain(datas):
    offset = 0x6A1
    cursor = offset
    while cursor < 0x1006A1:
        if datas[cursor: cursor +4] == b'\xff\xff\xff\xff':
            return cursor
        delta = bg(datas[cursor+4: cursor+8])
        cursor += 8 + delta
    input('Entree vide dans liste chainée non trouvée. Presser une touche pour quitter.')
    exit()

def moveArp(datas, arpdir, datasarp, od, nfrom, nto, dto):
    dirarp = arpdir[nfrom]
    arpdata = datasarp[nfrom]
    categorie = getCategorie(dirarp[0], datas, od)
    nc = nto.to_bytes(2, 'big')
    nc += nto.to_bytes(2, 'big')
    nc += categorie[4].to_bytes(1, 'big')
    nc += categorie[5].to_bytes(1, 'big')
    offsetCat = 0xA1 + (nto * 6)
    dto[offsetCat: offsetCat + len(nc)] = nc
    offsetLC = findOffsetListChain(dto)
    dto[offsetLC: offsetLC + len(arpdata)] = arpdata
    offsetDir = 0x1006A1 + (nto * 8)
    offsetLCrelatif = offsetLC - 0x6A1
    dto[offsetDir: offsetDir+4] = offsetLCrelatif.to_bytes(4,'big')
    dto[offsetDir+4: offsetDir+8] = dirarp[2].to_bytes(4,'big')

def moveArps(liseMove):
    listeMerge = expandStringList(liseMove, nbarp1, nbarp2)
    for (i, e) in enumerate(listeMerge):
        if e[0] == 0:
            datas = datas1
            arpdir=arpdir1
            datasarp=datasarp1
            od=od1
        else:
            datas = datas2
            arpdir=arpdir2
            datasarp=datasarp2
            od=od2
        nfrom=e[1]-1
        nto= i
        dto = datas3
        moveArp(datas, arpdir, datasarp, od, nfrom, nto, dto)
    return listeMerge

def readParams():
    # ex : "D:/Yamaha/Cle64GO/BACKUPS/PATARP/TESTS/5-8-F-13.X6G" "D:/Yamaha/Cle64GO/BACKUPS/PATARP/MYEDMEXP.X6G" "ARPS/TEST4.X6G" a8,a7,b34,b16,b6,b1,a1
    if len(sys.argv) > 4:
        n1 = sys.argv[1]
        n2 = sys.argv[2]
        n3 = sys.argv[3]
        l = sys.argv[4]
    else:
        print('Appel du programme : py moxfArpeges nomFichierA nomFichierB nomFichierDeSortie liste')
        print("nomFichierA = nom d'un fichier d'extension .X6G ou .X6A contenant des arpèges à fusionner")
        print("nomFichierB = nom d'un fichier d'extension .X6G ou .X6A contenant des arpèges à fusionner")
        print("nomFichierDeSortie = nom du fichier .X6G qui sera créer avec les arpèges fusionnées.")
        print('Les noms de fichiers doivent être entourés de guillemets (caractère " ) ')
        print("liste ex : a5, b34 signifie fusionner l'arpège numéro 5 du fichier A et l'arpège numéro 34 du fichier B")
        print("Les numéros d'arpèges sont compris entre 1 et 256")
        print("""Ex : py moxfArpeges "D:/Yamaha/Cle64GO/BACKUPS/PATARP/TESTS/5-8-F-13.X6G" "D:/Yamaha/Cle64GO/BACKUPS/PATARP/MYEDMEXP.X6G" "ARPS/TEST4.X6G" a8,a7,b34,b16,b6,b1,a1""")
        exit()

    (filepath, filename) = os.path.split(n1)
    (shortname, extension) = os.path.splitext(filename)
    if extension.upper() != '.X6G' and extension.upper() != '.X6A':
        input("Le fichier A doit être d'extension .X6G ou .X6A")
        exit()
    (filepath, filename) = os.path.split(n2)
    (shortname, extension) = os.path.splitext(filename)
    if extension.upper() != '.X6G' and extension.upper() != '.X6A':
        input("Le fichier B doit être d'extension .X6G ou .X6A")
        exit()
    (filepath, filename) = os.path.split(n3)
    (shortname, extension) = os.path.splitext(filename)
    if extension.upper() != '.X6G' :
        input("Le fichier de sortie doit être d'extension .X6G")
        exit()

    return n1, n2, n3, l


n1,n2,n3, l = readParams()
print(n1)
print(n2)
print(n3)
print(l)

file1 = open(n1,mode='rb')
datas1 = file1.read()
file1.close()

file2 = open(n2,mode='rb')
datas2 = file2.read()
file2.close()

od1 = findOffsetDARP(datas1)
arpdir1 = litDirArp(datas1, od1)
nbarp1 = len(arpdir1)
datasarp1 = getDatasArp(datas1, od1, arpdir1)
pg1()

od2 = findOffsetDARP(datas2)
arpdir2 = litDirArp(datas2, od2)
nbarp2 = len(arpdir2)
datasarp2 = getDatasArp(datas2, od2, arpdir2)
pg2()

writeinfo(n1, arpdir1, datasarp1, datas1, od1)
writeinfo(n2, arpdir2, datasarp2, datas2, od2)
datas3 = bytearray(makeEmptyArps())

lm = moveArps(l)

writeArp(n3, datas3)

file3 = open(n3,mode='rb')
datas3 = file3.read()
file3.close()

od3 = findOffsetDARP(datas3)
arpdir3 = litDirArp(datas3, od3)
nbarp3 = len(arpdir3)
datasarp3 = getDatasArp(datas3, od3, arpdir3)
pg3()

print('Terminé')
